(*
  The FITS package, version 1.0.
*)

(*
  This package can be used to read/write an image in
  FIST format; it was created by Luca Caucci. You
  may send any comment, suggestion or idea to
  caucci@cs.unibo.it.
*)

(*
  HISTORY:
    2002-07-18: Version 1.0 released
*)

(* needed to read/write binary data from/to a file *)
BeginPackage["FITS`", "Utilities`BinaryFiles`"]

readFITS::usage = "readFITS[fileName] opens a FITS file and, if
the file is in FITS format, returns its header and its data array.
The FITS file is supposed not to have any extension blocks
and the number of axes must be two, that is to say we are
handling a 2D image. The header is checked for validity and
if the image is 2D, the data array is read. This function
returns a list of two matrices; the former is the FITS header,
the latter the data array. The elements returned in the data
array will be real numbers in the range [0, 1]; the header
is a list of strings. If the FITS file is not valid, a list
of two empty matrices is returned and an error message is
issued.";

saveFITS::usage = "saveFITS[fileName, data] saves a 2D image
to a FITS file. The file name is passed to the function as a
string. The data must be a 2D real matrix whose elements are
in the range [0, 1]. The option DataType can be used to tell
data type and length and must be 8 or 16 or 32 or -32 or -64.
A positive value stands for integer numbers written to the
FITS file while a negative value stands for IEEE754 floating
point numbers. The absolute value of the DataType option indicates
the number of bits required by each element of the data matrix.";

DataType::usage = "DataType is an option for saveFITS. It
lets the user choose the type of the data written to the
FITS file. It can assume the values 8 for unsigned 8-bit
integer, 16 for signed 16-bit intger, 32 for 32-bit integer,
-32 for single-precision IEEE754 floating point and -64
for double-precision IEEE754 floating point.";

Begin["`Private`"]

(* error messages *)
FITS::PrematureEnd = "Premature end of FITS file encountered.";
FITS::CorruptedHeader = "Corrupted FITS header: missing `1` keyword.";
FITS::NonStandardHeader = "Non-standard FITS file.";
FITS::DataType = "`1` is not a valid data type specification.";
FITS::Not2DImage = "Not a 2-D image.";
FITS::BadDimension = "`1` is not a valid value for `2`.";
FITS::NoENDFound = "No END keyword found at the end of the header.";
FITS::InvalidCardImage = "Invalid card image.";
FITS::InvalidKeywordLength = "Invalid keyword length.";
FITS::InvalidCardImageLength = "Invalid card image length.";

(*
  This function, given the FITS header as an array
  of strings whose lenghts are 80 (that is to say as
  a matrix of characters), looks up for a given key.
  If the key is found, its index into the array of
  string is returned, otherwise the function returns
  zero.
*)
getHeaderEntryIndex[header_List, key_String] := 
  Module[{i, found, entry},
    i = 1;
    found = False;
    (* scan the array until either the key is
    found or the whole array has been scanned *)
    While[(i <= Length[header]) && !found,
      entry = header[[i]];
      If[StringTake[entry, {1, 8}] == key,
        found = True,
        i++;
      ]
    ];
    (* return ``0'' if not found *)
    If[found, i, 0]
  ]

(*
  This function, given a FITS header (i.e. an array
  of strings whose lengths are 80 characters), scans it
  looking for a given key, and if the key is found the
  corresponding entry (a string) is returned. If the
  key being searched for doesn't exist in the header
  an empty string is returned.
*)
getHeaderEntry[header_List, key_String] :=
  Module[{i},
    i = getHeaderEntryIndex[header, key];
    (* return an empty string if not found *)
    If[i != 0, header[[i]], ""]
  ]

(*
  This function opens a FITS file and, if the file is in
  FITS format, returns its header and its data array. The
  FITS file is supposed not to have any extension blocks
  and the number of axes must be two, that is to say we are
  handling a 2D image. The header is checked for validity and
  if the image is 2D, the data array is read. This function
  returns a list of two matrices; the former is the FITS header,
  the latter the data array. The elements returned in the data
  array will be real numbers in the range [0, 1]; the header
  is a list of strings. If the FITS file is not valid, a list
  of two empty matrices is returned and an error message is
  issued.
*)
readFITS[fileName_String] := 
  Module[{stream, header, tmpString, cardImage, cont, i,
    error, data, orizDim, vertDim, bitPix, dataMin, dataMax},
    header = data = {};
    cont = True;
    error = False;
    (* open the file *)
    stream = OpenReadBinary[fileName];
    If[Head[stream] == InputStream,
      While[cont,
        (* read a whole physical block *)
        tmpString = FromCharacterCode[ReadListBinary[stream, Byte, 2880]];
        If[StringLength[tmpString] != 2880,
          Message[FITS::PrematureEnd];
          error = True;
          cont = False,
          i = 0;
          (* convert the block into card images *)
          While[cont && (i < 36),
            cardImage = StringTake[tmpString, {i * 80 + 1, (i + 1) * 80}];
            AppendTo[header, cardImage];
            i++;
            (* check if the last card image is reached *)
            If[StringTake[cardImage, {1, 3}] == "END",
              cont = False;
            ]
          ]
        ]
      ];
      (* check if the header is valid *)
      If[!error && isAValidHeader[header],
        (* read the number of rows and columns *)
        orizDim = getEntryInteger[getHeaderEntry[header, "NAXIS1  "]];
        vertDim = getEntryInteger[getHeaderEntry[header, "NAXIS2  "]];
        (* read the data size and format *)
        bitPix = getEntryInteger[getHeaderEntry[header, "BITPIX  "]];
        (* read the data array, depending on the data size and format *)
        Switch[bitPix,
          8,   Do[AppendTo[data, ReadListBinary[stream, Byte, Evaluate[orizDim]]], {vertDim}],
          16,  Do[AppendTo[data, ReadListBinary[stream, SignedInt16, Evaluate[orizDim]]], {vertDim}],
          32,  Do[AppendTo[data, ReadListBinary[stream, SignedInt32, Evaluate[orizDim]]], {vertDim}],
          -32, Do[AppendTo[data, ReadListBinary[stream, Single, Evaluate[orizDim]]], {vertDim}],
          -64, Do[AppendTo[data, ReadListBinary[stream, Double, Evaluate[orizDim]]], {vertDim}]
        ];
        data = N[data];
        (* scale the data in the range [0, 1], real number *)
        dataMin = If[getHeaderEntryIndex[header, "DATAMIN "] != 0,
          getEntryFloatingPoint[header[[i]]],
          Min[data]
        ];
        dataMax = If[getHeaderEntryIndex[header, "DATAMAX "] != 0,
          getEntryFloatingPoint[header[[i]]],
          Max[data]
        ];
        (* avoid division-by-zero errors *)
        If[dataMax != dataMin,
          data = (data - dataMin) / (dataMax - dataMin);
        ];
        (* scale the data matrix based on the scale
        information specified in the FITS file *)
        i = getHeaderEntryIndex[header, "BSCALE  "];
        If[i != 0,
          data *= getEntryFloatingPoint[header[[i]]];
        ];
        i = getHeaderEntryIndex[header, "BZERO   "];
        If[i != 0,
          data += getEntryFloatingPoint[header[[i]]];
        ],
        (* if the file is not valid, return an
        empty header and an empty data matrix *)
        header = data = {};
      ];
      (* close the file *)
      Close[stream];
    ];
    (* return the header and the data matrix *)
    {header, data}
  ]
    
(*
  This function ckecks if the passed header is valid.
  Not only does it check whether the header obeys the
  FITS standard but it checks if the image is 2D.
*)
isAValidHeader[header_List] :=
  Module[{valid, num},
    valid = False;
    If[Length[header] !=  0,
      (* check the ``SIMPLE'' keyword *)
      If[StringTake[header[[1]], {1, 10}] == "SIMPLE  = ",
        (* check if the file adherers the standard *)
        If[getEntryBoolean[header[[1]]],
          (* check the ``BITPIX'' keyword *)
          If[StringTake[header[[2]], {1, 10}] == "BITPIX  = ",
            (* get the ``BITPIX'' value *)
            num = getEntryInteger[header[[2]]];
            (* check if it is valid *)
            If[Position[{8, 16, 32, -32, -64}, num] != {},
              (* check the ``NAXIS'' keyword *)
              If[StringTake[header[[3]], {1, 10}] == "NAXIS   = ",
                (* get the number of axes *)
                num = getEntryInteger[header[[3]]];
                (* check if the stored image is 2D *)
                If[num == 2,
                  (* get the number of columns *)
                  If[StringTake[header[[4]], {1, 10}] == "NAXIS1  = ",
                    num = getEntryInteger[header[[4]]];
                    If[num > 0,
                      (* get the number of rows *)
                      If[StringTake[header[[5]], {1, 10}] == "NAXIS2  = ",
                        num = getEntryInteger[header[[5]]];
                        If[num > 0,
                          (* check the last keyword *)
                          If[StringTake[header[[Length[header]]], {1, 3}] == "END",
                            (* everything is OK! *)
                            valid = True,
                            (* error messages *)
                            Message[FITS::NoENDFound];
                          ],
                          Message[FITS::BadDimension, num, "NAXIS2"];
                        ],
                        Message[FITS::CorruptedHeader, "NAXIS2"];
                      ],
                      Message[FITS::BadDimension, num, "NAXIS1"];
                    ],
                    Message[FITS::CorruptedHeader, "NAXIS1"];
                  ],
                  Message[FITS::Not2DImage];
                ],
                Message[FITS::CorruptedHeader, "NAXIS"];
              ],
              Message[FITS::DataType, num];
            ],
            Message[FITS::CorruptedHeader, "BITPIX"];
          ],
          Message[FITS::NonStandardHeader];
        ],
        Message[FITS::CorruptedHeader, "SIMPLE"];
      ]
    ];
    valid
  ]

(*
  This function gets the boolean value of a card image. 
  The card image is passed as a string of 80 characters.
*)
getEntryBoolean[entry_String] :=
  Module[{},
    If[StringLength[entry] == 80,
      StringTake[entry, {30}] == "T",
      (* invalid card image error *)
      Message[FITS::InvalidCardImage];
      Null
    ]
  ]

(*
  This function gets the integer value of a card image. 
  The card image is passed as a string of 80 characters.
*)
getEntryInteger[entry_String] :=
  Module[{tmpEntry, CommentPosition},
    If[StringLength[entry] == 80,
      tmpEntry = StringTake[entry, {11, 30}];
      (* check if a comment is present and if yes, remove it *)
      CommentPosition = StringPosition[tmpEntry, "/"];
      If[CommentPosition != {},
        (* discard the comment *)
        tmpEntry = StringTake[tmpEntry, CommentPosition[[1, 1]] - 1];
      ];
      Floor[ToExpression[tmpEntry]],
      (* invalid card image error *)
      Message[FITS::InvalidCardImage];
      Null
    ]
  ]

(*
  This function gets the integer value of a card image. 
  The card image is passed as a string of 80 characters.
*)
getEntryFloatingPoint[entry_String] :=
  Module[{tmpEntry, CommentPosition},
    If[StringLength[entry] == 80,
      tmpEntry = StringTake[entry, {11, 80}];
      (* check if a comment is present and if yes, remove it *)
      CommentPosition = StringPosition[tmpEntry, "/"];
      If[CommentPosition != {},
        tmpEntry = StringTake[tmpEntry, CommentPosition[[1, 1]] - 1];
      ];
      (* convert the exponential notation into a suitable format for Mathematica *)
      tmpEntry = StringReplace[tmpEntry, "D" -> "E"];
      tmpEntry = StringReplace[tmpEntry, "E" -> "* 10 ^"];
      ToExpression[tmpEntry],
      (* invalid card image error *)
      Message[FITS::InvalidCardImage];
      Null
    ]
  ]

(*
  This function gets the complex (both integer complex and
  real complex) value of a card image. The card image is
  passed as a string of 80 characters.
*)
getEntryComplex[entry_String] :=
  Module[{index1, index2, indexComma, tmpEntry, result, str1, str2, CommentPosition},
    result = Null;
    If[StringLength[entry] == 80,
      tmpEntry = StringTake[entry, {11, 80}];
      (* check if a comment is present and if yes, remove it *)
      CommentPosition = StringPosition[tmpEntry, "/"];
      If[CommentPosition != {},
        tmpEntry = StringTake[tmpEntry, CommentPosition[[1, 1]] - 1];
      ];
      (* find the starting and beginning indexes
      of the real part and the imaginary part *)
      index1 = StringPosition[entry, "("][[1, 1]];
      index2 = StringPosition[entry, ")"][[1, 1]];
      tmpEntry = StringTake[entry, {index1 + 1, index2 - 1}];
      indexComma = StringPosition[tmpEntry, ","];
      If[Length[indexComma] == 1,
        indexComma = indexComma[[1, 1]];
        (* take the real part and the imaginary part *)
        str1 = StringTake[tmpEntry, {1, indexComma - 1}];
        str2 = StringTake[tmpEntry, {indexComma + 1, StringLength[tmpEntry]}];
        (* convert the exponential notation into a suitable
        format for Mathematica's ``ToExpression'' function *)
        str1 = StringReplace[str1 , "D" -> "E"];
        str2 = StringReplace[str2 , "D" -> "E"];
        str1 = StringReplace[str1, "E" -> " * 10 ^"];
        str2 = StringReplace[str2, "E" -> " * 10 ^"];
        (* ``parse'' the string *)
        result = ToExpression[str1] + I * ToExpression[str2];
      ],
      (* invalid card image error *)
      Message[FITS::InvalidCardImage];
    ];
    result
  ]

(*
  This function gets the string value of a card image. 
  The card image is passed as a string of 80 characters.
*)
getEntryString[entry_String] :=
  Module[{indexes, result},
    If[StringLength[entry] == 80,
      result = "";
      indexes = StringPosition[entry, "'"];
      If[Length[indexes] >= 2,
        (* remove the delimiters *)
        result = StringTake[entry, {First[indexes][[1]] + 1, Last[indexes][[1]] - 1}];
        (* convert the escape sequence '' to ' *)
        StringReplace[result, "''" -> "'"];
      ],
      (* invalid card image error *)
      Message[FITS::InvalidCardImage];
    ];
    result
  ]

(*
  This function returns a new card image whose key is
  passed as first parameter and sets its boolean value
  to the second parameter passed to the function. The
  card image is returned as an 80-character string.
*)
setEntryBoolean[key_String, val_] :=
  Module[{result, charValue},
    result = "";
    If[StringLength[key] == 8,
      charValue = If[TrueQ[val], "T", "F"];
      result = StringJoin[key, "= ", Table[" ", {19}], charValue, Table[" ", {50}]],
      (* invalid keyword length error *)
      Message[FITS::InvalidKeywordLength];
    ];
    result
  ]

(*
  This function returns a new card image whose key is
  passed as first parameter and sets its integer value
  to the second parameter passed to the function. The
  card image is returned as an 80-character string.
*)
setEntryInteger[key_String, val_] :=
  Module[{result, integerValue, strLength},
    result = "";
    If[StringLength[key] == 8,
      (* do the one-dimensional and simplest string conversion *)
      integerValue = ToString[Ceiling[val], FormatType -> InputForm];
      strLength = StringLength[integerValue];
      If[strLength <= 20,
        result = StringJoin[key, "= ", Table[" ", {20 - strLength}], integerValue, Table[" ", {50}]],
        (* invalid card image length error *)
        Message[FITS::InvalidCardImageLength];
      ],
      (* invalid keyword length error *)
      Message[FITS::InvalidKeywordLength];
    ];
    result
  ]
    
(*
  This function returns a new card image whose key is
  passed as first parameter and sets its floating point
  value to the second parameter passed to the function.
  The card image is returned as an 80-character string.
*)
setEntryFloatingPoint[key_String, val_] :=
  Module[{result, realValue, strLength},
    result = "";
    If[StringLength[key] == 8,
      (* do the one-dimensional and simplest string conversion *)
      realValue = ToString[NumberForm[N[val], NumberFormat -> (SequenceForm[#1, "0D0", #3] &)]];
      strLength = StringLength[realValue];
      If[strLength <= 70,
        result = StringJoin[key, "= ", Table[" ", {70 - strLength}], realValue],
        (* invalid card image length error *)
        Message[FITS::InvalidCardImageLength];
      ],
      (* invalid keyword length error *)
      Message[FITS::InvalidKeywordLength];
    ];
    result
  ]
    
(*
  This function returns a new card image whose key is
  passed as first parameter and sets its integer complex
  value to the last two parameters passed to the function.
  The card image is returned as an 80-character string.
*)
setEntryComplexInteger[Key_String, realPart_Integer, imaginaryPart_Integer] :=
  Module[{result, integerRealValue, integerImaginaryValue, strLength, tmpString},
    result = "";
    If[StringLength[Key] == 8,
      (* do the one-dimensional and simplest string conversion *)
      integerRealValue = ToString[Ceiling[realPart], FormatType -> InputForm];
      integerImaginaryValue = ToString[Ceiling[imaginaryPart], FormatType -> InputForm];
      tmpString = StringJoin["(", integerRealValue, ",", integerImaginaryValue, ")"];
      strLength = StringLength[tmpString];
      If[strLength <= 70,
        result = StringJoin[Key, "= ", Table[" ", {70 - strLength}], tmpString],
        (* invalid card image length error *)
        Message[FITS::InvalidCardImageLength];
      ],
      (* invalid keyword length error *)
      Message[FITS::InvalidKeywordLength];
    ];
    result
  ]

(*
  This function returns a new card image whose key is
  passed as first parameter and sets its real complex
  value to the last two parameters passed to the function.
  The card image is returned as an 80-character string.
*)
setEntryComplexReal[Key_String, realPart_Real, imaginaryPart_Real] :=
  Module[{result, realRealValue, realImaginaryValue, strLength, tmpString},
    result = "";
    If[StringLength[Key] == 8,
      (* do the one-dimensional and simplest string conversion *)
      realRealValue = ToString[NumberForm[N[realPart],
        NumberFormat -> (SequenceForm[#1, "0D0", #3] &)]];
      realImaginaryValue = ToString[NumberForm[N[imaginaryPart],
        NumberFormat -> (SequenceForm[#1, "0D0", #3] &)]];
      tmpString = StringJoin["(", realRealValue, ",", realImaginaryValue, ")"];
      strLength = StringLength[tmpString];
      If[strLength <= 70,
        result = StringJoin[Key, "= ", Table[" ", {70 - strLength}], tmpString],
        (* invalid card image length error *)
        Message[FITS::InvalidCardImageLength];
      ],
      (* invalid keyword length error *)
      Message[FITS::InvalidKeywordLength];
    ];
    result
  ]
    
(*
  This function returns a new card image whose key is
  passed as first parameter and sets its string value
  to the second parameter passed to the function. The
  card image is returned as an 80-character string.
*)
setEntryString[key_String, val_] :=
  Module[{result, strLength, tmpStr},
    result = "";
    If[StringLength[key] == 8,
      (* we must pay attention in order to correctly
      handle strings such as "O'Hara" *)
      tmpStr = StringReplace[val , "'" -> "''"];
      (* encapsulate the string into a ''...'' pair *)
      tmpStr = StringJoin["'", tmpStr, "'"];
      strLength = StringLength[tmpStr];
      If[strLength <= 70,
        result = StringJoin[key, "= ", tmpStr, Table[" ", {70 - strLength}]],
        (* invalid card image length error *)
        Message[FITS::InvalidCardImageLength];
      ],
      (* invalid keyword length error *)
      Message[FITS::InvalidKeywordLength];
    ];
    result
  ]
    
(*
  This function returns a new card image whose key is
  passed as first parameter and fills the last 72
  characters of the card image with the second parameter
  passed to the function. The card image is returned as
  an 80-character string. This function can be used,
  for example, to set the comment field of a HISTORY or
  COMMENT image card.
*)
setEntryNoValue[key_String, val_] :=
  Module[{result, strLength},
    result = "";
    If[StringLength[key] == 8,
      strLength = StringLength[val];
      If[strLength <= 72,
        (* no `` = '' symbol required here! *)
        result = StringJoin[key, val, Table[" ", {72 - strLength}]],
        (* invalid card image length error *)
        Message[FITS::InvalidCardImageLength];
      ],
      (* invalid keyword length error *)
      Message[FITS::InvalidKeywordLength];
    ];
    result
  ]

(*
  This function saves a 2D image as a FITS file. The file
  name is passed to the function as a string. The data must
  be a real matrix whose elements are in the range [0, 1].
  The option DataType can be used to tell the data type and
  length and must be 8 or 16 or 32 or -32 or -64. A positive
  value stands for integer numbers written to the FITS file
  while a negative value stands for IEEE754 floating point
  numbers. The absolute value of the DataType option indicates
  the number of bits required by each element of the data
  matrix.
*)

(* set the default data rappresentation *)
Options[saveFITS] = {DataType -> 16};

saveFITS[fileName_String, data_:MatrixQ, dataOption___?OptionQ] :=
  Module[{header, stream, i, j, buff, dataLen},
    (* read the data type option *)
    {dataLen} = {DataType} /. {dataOption} /. Options[saveFITS];
    (* check the data type and length *)
    If[Position[{8, 16, 32, -32, -64}, dataLen] != {},
      (* build the FITS header *)
      header = {};
      (* since we correctly use some functions we have
      defined, we can avoid worrying about errors :) *)
      AppendTo[header, setEntryBoolean["SIMPLE  ", True]];
      AppendTo[header, setEntryInteger["BITPIX  ", dataLen]];
      AppendTo[header, setEntryInteger["NAXIS   ", 2]];
      AppendTo[header, setEntryInteger["NAXIS1  ", Dimensions[data][[2]]]];
      AppendTo[header, setEntryInteger["NAXIS2  ", Dimensions[data][[1]]]];
      (* Luca Caucci's ``signature''! *)
      AppendTo[header, setEntryNoValue["COMMENT ", "This file has been written using Mathematica's FITS package."]];
      AppendTo[header, setEntryNoValue["COMMENT ", "The Mathematica's FITS package was created by Luca Caucci."]];
      (* END image card *)
      AppendTo[header, setEntryNoValue["END     ", ""]];
      (* create a new file *)
      stream = OpenWriteBinary[fileName];
      For[i = 1, i <= Length[header], ++i,
        buff = ToCharacterCode[header[[i]]];
        (* write each card image *)
        WriteBinary[stream, buff, ByteConversion -> (ToBytes[#, IntegerConvert -> Int8] &)];
      ];
      (* check if the physical block should be completed. The size of a
        physical block is 2880 bytes, that is to say 36 card images. *)
      If[Mod[Length[header], 36] != 0,
        j = 36 - Mod[Length[header], 36];
        For[i = 1, i <= j, ++i,
          WriteBinary[stream, Table[ToCharacterCode[" "], {80}],
            ByteConversion -> (ToBytes[#, IntegerConvert -> Int8] &)];
        ];
      ];
      (* switch for each data type *)
      Switch[dataLen,
        (* 8 bits, unsigned integer *)
        8,
        For[i = 1, i <= Length[data], ++i,
          (* scale and write the data *)
          WriteBinary[stream, Floor[data[[i]] * (2^8 - 1)], ByteConversion -> (ToBytes[#, IntegerConvert -> Int8] &)];
        ];
        (* check if the physical block must be padded with zeros *)
        If[Mod[Apply[Times, Dimensions[data]] * 1, 2880] != 0,
          j = 2880 - Mod[Apply[Times, Dimensions[data]] * 1, 2880];
          WriteBinary[stream, Table[0, {j}], ByteConversion -> (ToBytes[#, IntegerConvert -> Int8] &)];
        ],
        (* 16 bits, signed integer *)
        16,
        For[i = 1, i <= Length[data], ++i,
          (* scale and write the data *)
          WriteBinary[stream, Floor[(data[[i]] - (1 / 2)) * (2^16 - 1)], ByteConversion -> (ToBytes[#, IntegerConvert -> Int16] &)];
        ];
        (* check if the physical block must be padded with zeros *)
        If[Mod[Apply[Times, Dimensions[data]] * 2, 2880] != 0,
          j = 2880 - Mod[Apply[Times, Dimensions[data]] * 2, 2880];
          WriteBinary[stream, Table[0, {j}], ByteConversion -> (ToBytes[#, IntegerConvert -> Int8] &)];
        ],
        (* 32 bits, signed integer *)
        32,
        For[i = 1, i <= Length[data], ++i,
          (* scale and write the data *)
          WriteBinary[stream, Floor[(data[[i]] - (1 / 2)) * (2^32 - 1)], ByteConversion -> (ToBytes[#, IntegerConvert -> Int32] &)];
        ];
        (* check if the physical block must be padded with zeros *)
        If[Mod[Apply[Times, Dimensions[data]] * 4, 2880] != 0,
          j = 2880 - Mod[Apply[Times, Dimensions[data]] * 4, 2880];
          WriteBinary[stream, Table[0, {j}], ByteConversion -> (ToBytes[#, IntegerConvert -> Int8] &)];
        ],
        (* 32 bits, IEEE754 single precision *)
        -32,
        For[i = 1, i <= Length[data], ++i,
          (* write the data *)
          WriteBinary[stream, N[data[[i]]], ByteConversion -> (ToBytes[#, RealConvert -> Single] &)];
        ];
        (* check if the physical block must be padded with zeros *)
        If[Mod[Apply[Times, Dimensions[data]] * 4, 2880] != 0,
          j = 2880 - Mod[Apply[Times, Dimensions[data]] * 4, 2880];
          WriteBinary[stream, Table[0, {j}], ByteConversion -> (ToBytes[#, IntegerConvert -> Int8] &)];
        ],
        (* 64 bits, IEEE754 double precision *)
        -64,
        For[i = 1, i <= Length[data], ++i,
          (* write the data *)
          WriteBinary[stream, N[data[[i]]], ByteConversion -> (ToBytes[#, RealConvert -> Double] &)];
        ];
        (* check if the physical block must be padded with zeros *)
        If[Mod[Apply[Times, Dimensions[data]] * 8, 2880] != 0,
          j = 2880 - Mod[Apply[Times, Dimensions[data]] * 8, 2880];
          WriteBinary[stream, Table[0, {j}], ByteConversion -> (ToBytes[#, IntegerConvert -> Int8] &)];
        ]
      ];
      (* close the file *)
      Close[stream],
      (* invalid data type specification *)
      Message[FITS::DataType, dataLen];
    ];
  ]
    
End[]

EndPackage[]
